#ifndef SECURE_CHANNEL_PEER_H
#define SECURE_CHANNEL_PEER_H

#include <string>

/* The pre-shared key for part 1. */
extern const char PRE_SHARED_KEY[16];

/**
 * A peer in our secure channel scheme.  TODO: You need to implement
 * subclasses of this class that carry out each part of the project.
 */
class SecureChannelPeer {
public:
  /* Set the peer's socket object and whether or not the peer is a server. */
  SecureChannelPeer(int sock, bool is_server)
    : sock(sock), is_server(is_server) {}

  virtual ~SecureChannelPeer() {}

  /**
   * Negotiate a session key with the server and do any necessary
   * authentication. This method will only be called once and only if
   * this node is the client. The server should simply handle as many
   * initial received messages as necessary as initiation messages. If
   * the protocol requires multiple messages for authentication, you
   * must track that internally and handle the received messages
   * through ProcessReceivedMessage.
   */
  virtual void InitiateSession() {}

  /**
   * Send a message, encrypted if necessary.
   */
  virtual void SendEncryptedMessage(const std::string &message) {}

  /**
   * Process a received message, decrypting and checking MAC as
   * necessary. Send any replies necessitated by the protocol.
   */
  virtual void ProcessReceivedMessage(const std::string &message) {}
protected:
  int sock;
  bool is_server;
};


/* TODO: You should implement these functions to return an instance of
 * your Peer classes for each part of the project. A well-designed
 * implementation design may be able to reuse the same code for the
 * client and the server. */
SecureChannelPeer *
GetCleartextPeer(int sock, bool is_server); /* Part 0 */
SecureChannelPeer *
GetPreSharedKeyPeer(int sock, bool is_server); /* Part 1 */
SecureChannelPeer *
GetKeyNegotiatingPeer(int sock, bool is_server); /* Part 2 */
SecureChannelPeer *
GetAuthenticatedKeyNegotiatingPeer(int sock, bool is_server); /* Part 3 */

#endif /* SECURE_CHANNEL_PEER_H */
