#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <fstream>
#include <string>
#include <iostream>

#include <arpa/inet.h>
#include <sys/types.h>
#include <inttypes.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <unistd.h>

#include "SecureChannelPeer.h"

using std::cout;
using std::cerr; using std::cin; using std::endl;
using std::ofstream;
using std::exit;
using std::fprintf;
using std::string;
using std::strncmp;

bool logging_enabled_g = false;
std::string logfile_g = "network_traffic.out";


const char PRE_SHARED_KEY[16] = {'\xe1', 'c', '5', '\x1a', '\xa1', '\x13',
                                 't', '\xc5','z', '\xa8', '2', 'C', '\n',
                                 '\xb6', ')', '9'};

std::string ReceiveMessage(int sock) {
  uint16_t msglen;
  int res = recv(sock, &msglen, sizeof(msglen), 0);

  if (!res) {
    return "";
  }

  if (res != sizeof(msglen)) {
    goto ERROR;
  }
  msglen = ntohs(msglen);

  if (!msglen) {
    fprintf(stderr, "Received empty message (invalid)! Exiting.\n");
    exit(4);
  }

  char buf[msglen];
  if (recv(sock, buf, msglen, MSG_WAITALL) != msglen) {
    goto ERROR;
  }

  if(logging_enabled_g){
      ofstream fout;
      fout.open(logfile_g.c_str(), std::ios::app);
      fout <<"Received message: "<<endl;
      fout <<std::string(buf, msglen);
      fout<<endl<<endl;
      fout.close();
  }
  fprintf(stderr, "Received message with length %hd\n", msglen);
  return std::string(buf, msglen);
ERROR:
  fprintf(stderr, "Error receving message! Exiting.\n");
  exit(3);
}

void SendMessage(int sock, const std::string &msg) {
  char buf[2 + msg.size()];
  *((uint16_t *)buf) = htons(msg.size());
  memcpy(buf + 2, msg.data(), msg.size());

  int lefttosend = sizeof(buf);
  int sent = 0;
  int res = 0;
  while (lefttosend) {
    res = send(sock, buf + sent, lefttosend, 0);
    if (res == -1) {
      fprintf(stderr, "Error sending message! Exiting.\n");
      exit(4);
    }
    sent += res;
    lefttosend -= res;
  }

  if(logging_enabled_g){
      ofstream fout;
      fout.open(logfile_g.c_str(), std::ios::app);
      fout <<"Sending message: "<<endl;
      fout <<msg;
      fout <<endl<<endl;
      fout.close();
  }
  fprintf(stderr, "Sent message with length %u\n", msg.size());
}


SecureChannelPeer *GetPeer(int peer_type, int sock, bool is_server) {
 switch(peer_type) {
  case 0:
    return GetCleartextPeer(sock, is_server);
  case 1:
    return GetPreSharedKeyPeer(sock, is_server);
  case 2:
    return GetKeyNegotiatingPeer(sock, is_server);
  case 3:
    return GetAuthenticatedKeyNegotiatingPeer(sock, is_server);
  default:
    fprintf(stderr, "Bad part number %d!\n", peer_type);
    exit(1);
  }
}


void main_loop(int sock, SecureChannelPeer *peer) {
  fd_set readfds;
  for (;;) {
    FD_ZERO(&readfds);
    FD_SET(0, &readfds);
    FD_SET(sock, &readfds);
    if (select(sock+1, &readfds, NULL, NULL, NULL) == -1) {
      perror("Error in select");
      exit(1);
    }

    if (FD_ISSET(sock, &readfds)) {
      string lastMessage = ReceiveMessage(sock);
      if (lastMessage.empty()) {
        break;
      } else {
      }
      peer->ProcessReceivedMessage(lastMessage);
    }

    if (FD_ISSET(0, &readfds)) {
      string to_send;
      getline(cin, to_send);
      peer->SendEncryptedMessage(to_send);
   }

  }
}

void parse_args(int argc, char** argv,
                int* port,
                int* peer_type,
                bool* is_server)
{
    if(argc == 1){
        return;
    }
    char c;
    while ((c = getopt(argc, argv, "L:p:P:")) != -1)
    {
        switch(c)
        {
             case 'L':
                logging_enabled_g = true;
                logfile_g = optarg;
                break;
             case 'p':
                *is_server = false;
                 *port = atoi(optarg);
                 break;
             case 'P':
                 *peer_type = atoi(optarg);
                 break;
          }
    }
}

int main(int argc, char **argv) {

  int sock = socket(AF_INET, SOCK_STREAM, 0);
  int port = 0;
  int peer_type = 0;
  bool is_server = true;
  logging_enabled_g = false;
  parse_args(argc, argv, &port,
             &peer_type,
             &is_server); 
  struct sockaddr_in addr;
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = INADDR_ANY;
  addr.sin_port = htons(0);
  if (bind(sock, (struct sockaddr*)&addr, sizeof(addr)) == -1) {
    fprintf(stderr, "Couldn't bind to port %d!\n", port);
    exit(2);
  }
    
  // Print the port if we're the server.
  if (is_server) {
    socklen_t len = sizeof(addr);
    getsockname(sock, (struct sockaddr *)&addr, &len);
    printf("Using port %d\n", ntohs(addr.sin_port));
    fflush(stdout);
  }

  if (is_server) { // Server.
    if (listen(sock, 5) == -1) {
      fprintf(stderr, "Couldn't listen!\n");
      exit(1);
    }

    for (;;) {
      int conn = accept(sock, NULL, NULL);
      SecureChannelPeer *peer = GetPeer(peer_type, conn, is_server);
      cerr << "DEBUG: server accepted connection" << endl;
      main_loop(conn, peer);
      close(conn);
    }
  } else { // Client.
    addr.sin_port = htons(port);
    if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
      perror("Couldn't connect to server");
      exit(1);
    }
    SecureChannelPeer *peer = GetPeer(peer_type, sock, is_server);

    peer->InitiateSession();
    main_loop(sock, peer);
  }

  puts("All done, exiting successfully.");
  return 0;
}

